package net.sergeych.crypto2

import com.ionspin.kotlin.crypto.util.encodeToUByteArray
import kotlinx.serialization.Serializable
import kotlinx.serialization.Transient
import net.sergeych.bipack.BipackEncoder
import net.sergeych.bipack.Unsigned
import net.sergeych.synctools.ProtectedOp
import net.sergeych.synctools.invoke

object PBKD {

    @Serializable
    class Params(
        val kdf: KDF,
        @Unsigned
        val startOffset: Int,
        @Transient
        private var precalculatedKey: SymmetricKey?=null
    ) {

        /**
         * Derive a key from the password in the idempotent manner.
         * If the [keyId] is provided, it will be used to determine key type (otherwise default asymmetric)
         * and to check the result, and the exception will be thrown.
         *
         * Note that it is reasonably effective to use with linked keys restoring (keys based on the same password).
         * It can also be called repeatedly.
         *
         * @param password to derive from
         * @param keyId optional key id, to get a key type from and check the result
         * @throws IncorrectPasswordException if [keyId] is provided and does not match the derivation result
         */
        fun derive(password: String, keyId: KeyId? = null): SymmetricKey {
            // could be already calculated
            precalculatedKey?.let { return it }

            // check key id sanity
            keyId?.let {
                if (it.id.magick != KeysMagickNumber.defaultSymmetric.number) {
                    throw NotSupportedException("deriving key of type ${it.id.magick} is not implemented")
                }
            }
            // this will not be ok when we support other key types:
            val keySize = SymmetricKey.keyLength

            // derive
            val bytes = generate(kdf, password)
            val key = SymmetricKey(bytes.sliceArray(startOffset..<startOffset + keySize), this)

            // check if we can checking
            if( keyId != null && keyId != key.id )
                throw IncorrectPasswordException()

            // save for later
            precalculatedKey = key

            return key
        }
    }

    /**
     * Create [PBKD.Params] to derive several keys from the single password
     */
    fun createMutlipleParams(complexity: KDF.Complexity,keysCount: Int): List<Params> {
        TODO()
    }

    private val op = ProtectedOp()

    private class Entry(val kdf: KDF) {
        val op = ProtectedOp()
        var data: UByteArray? = null
    }

    /**
     * Binary id is a binary hash-capable key, we use it. byte array can't be a hash key in kotlin: its value
     * does not depend on the byte content.
     */
    fun key(kdf: KDF, password: String): BinaryId =
        BinaryId.createFromBytes(
            0,
            BipackEncoder.encode(kdf) + blake2b3l(password.encodeToUByteArray()).toByteArray()
        )

    private val cache = HashMap<BinaryId, Entry>()

    /**
     * Get a plain bytes by the KDF parameters specified. It implements short-term caching of
     * KDF results this could be called repeatedly.
     *
     * When called from different threads, allows deriving in parallel, but this could deplete memory
     * or CPU cores.
     */
    fun generate(kdf: KDF, password: String): UByteArray {
        val entry = op.invoke {
            cache.getOrPut(key(kdf, password)) { Entry(kdf) }
        }
        return entry.op {
            if (entry.data == null) {
                entry.data = entry.kdf.derive(password)
            }
            entry.data!!
        }
    }
}



